// VRdisplay.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2014 P. Skands
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Vincia ROOT utilities

#ifndef VINCIAROOTDISPLAY_H
#define VINCIAROOTDISPLAY_H

// Include the VINCIA ROOT headers 
// (ROOT and STDLIB headers included in VRsettings.h)

#include "VRfile.h"
#include "VRgraph.h"
#include "VRsettings.h"

// Put all the rest in Vincia namespace

namespace Vincia {
  
  // ------------------------------------------------------------------------
  // class VRdisplay
  // Contains the base class for a single 1D plot

  class VRdisplay {
    
  public:

    // Constructor
    VRdisplay(string displayNameIn, int xSizIn = 300, int ySizIn = 400,
              int xOffsetIn=-1, int yOffsetIn=-1) 
      : nPad(1), isInitPad(false), hasAxesSav(false), has2d(false), 
      hasDataSav(false), hasTitle(false), 
      hasLegend(false), hasLegendChi2(false), showLegendChi2(false),
      hasLegendUncVar(false), hasMCnames(false), hasDataRef(false), 
      hasLeftLabel(false), hasRightLabel(false), hasTopBox(false), 
      hasLeftLabelPtr(false), hasRightLabelPtr(false),
      rebinLikeData(false), zeroSkipping(false), hasSignature(false) {
      displayNameSav   = displayNameIn;
      displayCanvasPtr = new TCanvas(displayNameIn.c_str(),
                                     displayNameIn.c_str(), 
                                     xOffsetIn, yOffsetIn,xSizIn,ySizIn);      
    }
  
    // Init pointers and set settings
    void init(VRsettings* settingsPtrIn);

    // Functions to Get    
    bool getLogY() { return logY; }
    bool getLinY() { return !logY; }
    bool getLogX() { return logX; }
    bool getLinX() { return !logX; }
    bool getNormalize() { return normalize; }
    int  getOptStat() { return optStat; }
    double getSigmaMC() { return sigmaStat; }    
    double getNormFac() { return normFac; }
    string getName() { return displayNameSav; }

    // Functions to check whether axes and data have been defined
    bool hasAxes() {return hasAxesSav;}
    bool hasData() {return hasDataSav;}    

    // Add pad (for ratio to data or decomposition of vector of histos)
    void addPad();

    // Goto pad
    void gotoPad(int iPadIn = 1);

    // Functions to Set
    void setVerbose(int verboseIn) {verbose = verboseIn;};
    void setOptStat(int optStatIn) {optStat = optStatIn;}
    void setSigmaMC(double sigmaStatIn) {sigmaStat = sigmaStatIn;};
    void setNormalize(bool normalizeIn = true, double normFacIn = 1.0) {
      normalize  = normalizeIn;
      normFac    = normFacIn;
    }
    void setLogY(bool logYin = true) { 
      logY = logYin;
      displayCanvasPtr->cd(1); 
      if (logYin) gPad->SetLogy();
      else gPad->SetLogy(0);
      gPad->cd(1);
      if (logYin) gPad->SetLogy(); 
      else gPad->SetLogy(0);
    }
    void setLinY(bool linYin = true) { setLogY(!linYin); }
    void setLogX(bool logXin = true) { 
      logX = logXin;
      gPad->cd(1);
      if (logXin) gPad->SetLogx(); 
      else gPad->SetLogx(0);
    }
    void setLinX(bool linXin = true) { setLogX(!linXin); }
    void setSignature(string signatureIn="V I N C I A R O O T") {
      signature = signatureIn;
    }
    void setLeftLegend(bool leftLegendIn=true) { leftLegend = leftLegendIn; } 
    void setShowRatioPad(bool showRatioIn=true) { 
      showRatioPad = showRatioIn; 
    }
    void setRatioRange(double ratioMin, double ratioMax) {
      if (!hasAxes()) return;
      axesNormPtr->SetMinimum(ratioMin);
      axesNormPtr->SetMaximum(ratioMax);
      // Make copy of x axis for overdrawing 
      int nBin = axesNormPtr->GetNbinsX();
      delete xCtrPtr;
      delete xUppPtr;
      xCtrPtr = new TLine(axesNormPtr->GetBinLowEdge(1),1.0,
                          axesNormPtr->GetBinLowEdge(nBin+1),1.0);
      xUppPtr = new TLine(axesNormPtr->GetBinLowEdge(1),ratioMax,
                          axesNormPtr->GetBinLowEdge(nBin+1),ratioMax);
    }    
    void setShowRelUnc(bool showRelUncIn=true) {
      showRelUncPads = showRelUncIn;
    }

    // Check if axes and/or data has been defined
    void setAxes(TH1D* initHis);

    // Set specific color scheme for specific graph
    // Return false if graph does not exist or color scheme not defined
    bool setColorScheme(int iScheme, int iGraph=1) {      
      int nGraphs    = graphs.size();
      int nGraphsHis = graphsHis.size();
      int nGraphsTot = nGraphs+nGraphsHis; 
      // Set color scheme for data
      if (iGraph == 0) {
        return dataPtr->setColorScheme(iScheme);
      } 
      // Set color schem for histograms
      else if (iGraph <= nGraphs) {        
        return graphs[iGraph-1].setColorScheme(iScheme);      
      }
      else if (iGraph <= nGraphsTot) {
        return graphsHis[iGraph-1-nGraphs].setColorScheme(iScheme);
      }
      else {
        return false;
      }
      return true;
    }

    //------------------------------------------------------------------
    // VRdisplay: TH1D I/O

    // Add a histogram, not necessarily the first
    void addTH1DPtr(TH1D* hisPtrIn) {
      if (!hasAxes()) {
        setAxes(hisPtrIn);
        graphs.resize(0);
      }
      graphs.push_back(VRgraph(hisPtrIn));      
      // Default: assign color set n to graph number n
      graphs[graphs.size()-1].setColorScheme(graphs.size()+graphsHis.size());
      if (graphs.size() >= 2 && showRatioPad && nPad <= 1) addPad();
    }

    // Add a 2D histogram (replaces any other histograms)
    void setTH2DPtr(TH2D* hisPtrIn) {
      // Set TH2D pointer
      th2dPtr = hisPtrIn;
      has2d   = true;
      // Switch off stats options
      th2dPtr->SetStats(optStat);
      // Reset all 1D content
      graphs.resize(0);
      nPad = 1;
      //
      displayCanvasPtr->cd();              
      // If several pads, first do main plot
      gotoPad(1);
      // Make sure we have room to plot Z (colour) scale
      gPad->SetRightMargin(0.145);
      gPad->SetBottomMargin(0.15);
      gPad->SetTopMargin(0.1);
      gPad->SetLeftMargin(0.125);
      // Set color scale
      const Int_t NRGBs = 6;
      const Int_t NCont = 20;      
      Int_t MyPalette[NCont];
      Double_t r[NRGBs]    = {0., 0.2, 0.4, 0.6, 1.0, 1.0};
      Double_t g[NRGBs]    = {0., 0.0, 0.4, 0.6, 1.0, 0.7};
      Double_t b[NRGBs]    = {0., 0.7, 0.4, 0.6, 0.2, 0.7};
      Double_t stop[NRGBs] = {0., 0.25, 0.4, .58, 0.7, 1.0};
      Int_t FI = TColor::CreateGradientColorTable(NRGBs, stop, r, g, b, NCont);
      for (int i=0;i<NCont;i++) MyPalette[i] = FI+i;
      MyPalette[0]=0;
      gStyle->SetPalette(NCont,MyPalette);    
      gStyle->SetNumberContours(NCont);    
    }

    // Add a histogram (with vector uncertainties), not necessarily the first
    void addTH1DPtr(vector<TH1D*> hisPtrIn) {
      if (!hasAxes()) {
        setAxes(hisPtrIn[0]);
        graphs.resize(0);
      }
      graphs.push_back(VRgraph(hisPtrIn));
      graphs[graphs.size()-1].setColorScheme(graphs.size()+graphsHis.size());
      
      // Add pad for ratios
      if (graphsHis.size() + graphs.size() >= 2 && nPad <= 1 
          && showRatioPad) addPad();

      // Add pad for decomposition of vector band
      if (hisPtrIn.size() >= 2 && nPad <= 2 && showRelUncPads) addPad();

    }
    
    // Get histogram pointer corresponding to graph i, variation j
    TH1D* getTH1DPtr(int i=0, int j=0) {
      if ( i <= int(graphs.size())-1 ) return graphs[i].getTH1DPtr(j);
      else return NULL;
    }

    //========================================================================
    // VRdisplay: FILE I/O

    //------------------------------------------------------------------------
    // getFormat(): 
    // Function to convert a list of integers to vector<int> 
    void getFormat(vector<int>& format, int i1, int i2, int i3=-1, int i4=-1,
                   int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, 
                   int i10=-1, int i11=-1, int i12=-1) {
      // Store file format 
      format.push_back(i1);
      format.push_back(i2); 
      if (i3 >= 0) format.push_back(i3); else return;
      if (i4 >= 0) format.push_back(i4); else return;
      if (i5 >= 0) format.push_back(i5); else return;
      if (i6 >= 0) format.push_back(i6); else return;
      if (i7 >= 0) format.push_back(i7); else return;
      if (i8 >= 0) format.push_back(i8); else return;
      if (i9 >= 0) format.push_back(i9); else return;
      if (i10 >= 0) format.push_back(i10); else return;
      if (i11 >= 0) format.push_back(i11); else return;
      if (i12 >= 0) format.push_back(i12); else return;
    }

    //------------------------------------------------------------------------
    // setDataFile():
    // Set file to be treated as experimental data 
    bool setDataFile(string dataFileName, bool hasSysErrors = true) {
      if (hasSysErrors) 
        return setDataFile(dataFileName,10,20,30,40,51,52,61,62);
      else return setDataFile(dataFileName,10,20,30,40,51,52);
    }
    bool setDataFile(string dataFileName, int i1, int i2, int i3=-1, int i4=-1, 
                     int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, 
                     int i10=-1, int i11=-1, int i12=-1);

    //------------------------------------------------------------------------
    // addHisFile() : 
    // Add file with saved histogram
    // (Default is to ignore any additional systematic errors)
    bool addHisFile(string hisFileName, bool hasSysErrors = false) {
      if (hasSysErrors) return addHisFile(hisFileName,10,20,30,40,51,52,61,62);
      else return addHisFile(hisFileName,10,20,30,40,51,52);
    }
    bool addHisFile(string hisFileName, int i1, int i2, int i3=-1, int i4=-1, 
                    int i5=-1, int i6=-1, int i7=-1, int i8=-1, int i9=-1, 
                    int i10=-1, int i11=-1, int i12=-1);

    //------------------------------------------------------------------------
    // Add various kinds of objects to display
    // (start with simple line. Can maybe be upgraded to general objects?)
    void addTLine(TLine* newLine) {
      linesPtr.push_back(newLine);
    }


    //------------------------------------------------------------------------
    // update(): 
    // Update Display 
    void update();

    //------------------------------------------------------------------------
    // saveDisplay():
    // Save display image to graphics file
    void saveDisplay(string prefix="none",  
                     string fileType="png",string option="Landscape");

    //------------------------------------------------------------------------
    // saveHistogram():
    // save main theory histogram to data file (in standard column format)
    // Note: extended format allows theory uncertainty variations to be included
    bool saveHistogram(string prefix="none", string ext="dat");

    //------------------------------------------------------------------------
    // printChi2()
    // Print out chi2 values, first of MC curves, then of saved histograms
    void printChi2();

    //========================================================================
    // VRdisplay: Plot titles, legends, descriptions

    //------------------------------------------------------------------------
    // setTitle():
    // Set title
    void setTitle(); 

    //------------------------------------------------------------------------
    // setMCnames():
    // Set MC name title
    void setMCnames(string MCnamesIn);

    //------------------------------------------------------------------------
    // setDataRef():
    // Set Data reference title
    void setDataRef(string dataRefIn);

    //------------------------------------------------------------------------
    // setLegend():
    // Set Legend 
    // Optionally force alignment: align = -1(left), 0(auto), 1(right)
    void setLegend(vector<string> legendIn, int alignIn=0); 
    void setLegend(vector<string> legendIn, double xLeg, 
                   double xWid=0.33, double dy=0.056, double yLegTop=0.82);

    //------------------------------------------------------------------------
    // setLegendChi2():
    // Set optional extra Legend containing chi2 values
    void setShowLegendChi2(bool showChi2 = true) {
      showLegendChi2 = showChi2;
    }
    
    //------------------------------------------------------------------------
    // setLegendUncVar():
    // Set Legend for uncertainty bands
    void setLegendUncVar(vector<string> legendIn); 

    //------------------------------------------------------------------------
    // setLeftLabel(), setRightLabel() 
    // Set Labels for beam/energy type (top left and right white-on-black) 
    void setLeftLabel(string leftLabelIn) { 
      hasLeftLabel = true;
      leftLabel = leftLabelIn; 
    }
    void setRightLabel(string rightLabelIn) {
      hasRightLabel = true;
      rightLabel = rightLabelIn;
    }

    // Set rebinning properties
    void setRebinLikeData(bool rebinLikeDataIn=true) {
      rebinLikeData = rebinLikeDataIn;     
      // Check if rebinning histograms to same binning as data
      if (hasDataSav && dataPtr!= 0 && rebinLikeData) {
        for (int iGraph=0; iGraph<int(graphs.size()); ++iGraph)
          graphs[iGraph].setRebinAxis(dataPtr,zeroSkipping);
        for (int iGraph=0; iGraph<int(graphsHis.size()); ++iGraph)
          graphsHis[iGraph].setRebinAxis(dataPtr,zeroSkipping);
      }      
    }
    void setZeroSkipping(bool zeroSkippingIn=true) {
      zeroSkipping  = zeroSkippingIn;
      for (int iGraph=0; iGraph<int(graphs.size()); ++iGraph)
        graphs[iGraph].setZeroSkipping(zeroSkippingIn);
      for (int iGraph=0; iGraph<int(graphsHis.size()); ++iGraph)
        graphsHis[iGraph].setZeroSkipping(zeroSkippingIn);
    }

  private:

    //========================================================================
    // VRdisplay: private data members and functions
    
    // Canvas
    string displayNameSav;
    TCanvas* displayCanvasPtr;

    // Internal variables
    bool logY, logX, normalize;
    bool showRatioPad, showRelUncPads;
    int verbose, optStat;
    double sigmaStat, normFac;

    // Input containers
    VRfile* dataFilePtr;
    vector<VRfile*> hisFilePtrs;
    
    // Global settings
    VRsettings* settingsPtr;

    // PADS
    int nPad;
    bool isInitPad;

    // PLOT

    // Use histogram to define axes
    TH1D* axesPtr;    
    TH1D* axesNormPtr;
    TH1D* axesVarPtr;
    TLine* xCtrPtr;
    TLine* xUppPtr;
    bool hasAxesSav;
    // Optional possibility to use display for 2D histograms
    TH2D* th2dPtr;
    bool has2d;

    // MC Graphs 
    vector<VRgraph> graphs;
    vector<VRgraphNorm> graphsNorm;

    // His graphs (for comparison to saved data files)
    vector<VRgraphHis> graphsHis;
    vector<VRgraphNorm> graphsHisNorm;

    // RelUnc pad
    vector<VRgraphVar> graphsVar;

    // Data graph (and normlaized one- and two-sigma variants)
    VRgraphData* dataPtr;
    VRgraphData* dataTwoSigmaPtr;
    VRgraphNorm* dataNormPtr;
    VRgraphNorm* dataTwoSigmaNormPtr;
    bool hasDataSav;

    // Plot title 
    TLatex* titleTexPtr;
    bool hasTitle;

    // Objects
    vector<TLine*> linesPtr;
    
    // LEGEND

    // Main Legend 
    TLegend* legendPtr;
    bool hasLegend;
    bool leftLegend;
    vector<string> legendSav;
    int alignSav;
    double xLegSav, xWidSav, dySav, yLegTopSav;

    // Chi2 Legend
    TLegend* legendChi2Ptr;
    bool hasLegendChi2, showLegendChi2;

    // Uncertainty band variation legends
    TLegend* legendUncVarTopPtr;
    TLegend* legendUncVarBotPtr;
    bool hasLegendUncVar;    

    // MC names and versions
    TLatex* MCnamesPtr;
    bool hasMCnames;
    string MCnamesSav;
    
    // Data reference
    TLatex* dataRefPtr;
    bool hasDataRef;

    // Left and right labels
    string leftLabel, rightLabel;
    TLatex* leftLabelPtr;
    TLatex* rightLabelPtr;
    bool hasLeftLabel, hasRightLabel;
    bool hasTopBox, hasLeftLabelPtr, hasRightLabelPtr;
    TPave* topBoxPtr;

    // Rebinning properties
    bool rebinLikeData, zeroSkipping;

    // VinciaRoot Signature (on lower right-hand edge of plots)
    TLatex* signaturePtr;
    bool hasSignature;
    string signature;

  };

}

#endif
